<?php

namespace Drupal\webform_number_limit\Plugin\WebformHandler;

use Drupal\Core\Form\FormStateInterface;
use Drupal\webform\WebformSubmissionInterface;
use Drupal\webform_options_limit\Plugin\WebformHandler\OptionsLimitWebformHandler;
use Drupal\webform_options_limit\Plugin\WebformOptionsLimitHandlerInterface;

/**
 * Webform handler that limits the number of submissions for options/checkboxes.
 *
 * @WebformHandler(
 *   id = "options_number_limit",
 *   label = @Translation("Options/Checkbox with number limit"),
 *   category = @Translation("Options"),
 *   description = @Translation("Define options/checkbox with number submission limits."),
 *   cardinality = \Drupal\webform\Plugin\WebformHandlerInterface::CARDINALITY_UNLIMITED,
 *   results = \Drupal\webform\Plugin\WebformHandlerInterface::RESULTS_IGNORED,
 *   submission = \Drupal\webform\Plugin\WebformHandlerInterface::SUBMISSION_REQUIRED,
 * )
 */
class OptionsNumberLimitWebformHandler extends OptionsLimitWebformHandler {

  use NumberLimitWebformHandlerTrait;

  /**
   * The form state.
   *
   * @var \Drupal\Core\Form\FormStateInterface
   */
  protected $formState;

  /**
   * Ugly workaround to get access to form values in getLimitInformation().
   *
   * {@inheritdoc}
   */
  public function alterElement(array &$element, FormStateInterface $form_state, array $context) {
    $this->formState = $form_state;
    parent::alterElement($element, $form_state, $context);
  }

  /**
   * Ugly workaround to get access to form values in getLimitInformation().
   *
   * {@inheritdoc}
   */
  public function validateOptionsElement(array $element, FormStateInterface $form_state) {
    $this->formState = $form_state;
    parent::validateOptionsElement($element, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'number_element_key' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function getSummary(): array {
    $summary = parent::getSummary();
    $settings = $this->getSettings();

    $element = $this->getWebform()->getElement($settings['number_element_key']);
    if ($element) {
      $webform_element = $this->elementManager->getElementInstance($element);
      $t_args = [
        '@title' => $webform_element->getAdminLabel($element),
        '@type' => $webform_element->getPluginLabel(),
      ];
      $summary['#settings']['number_element_key'] = $this->t('@title (@type)', $t_args);
    }
    elseif (empty($settings['number_element_key'])) {
      $summary['#settings']['number_element_key'] = [
        '#type' => 'link',
        '#title' => $this->t('Please add a new number element.'),
        '#url' => $this->getWebform()->toUrl('edit-form'),
      ];
    }
    else {
      $summary['#settings']['number_element_key'] = [
        '#markup' => $this->t("'@element_key' is missing.", ['@element_key' => $settings['number_element_key']]),
        '#prefix' => '<b class="color-error">',
        '#suffix' => '</b>',
      ];
    }

    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    // Number element settings.
    $form['number_element_settings'] = [
      '#type' => 'details',
      '#title' => $this->t('Number element settings'),
      '#open' => TRUE,
      '#weight' => -1,
    ];

    $numberElements = $this->getNumberElements();
    $numberElement = $this->configuration['number_element_key'];

    if ($numberElement && !isset($numberElements[$numberElement])) {
      $this->showMissingElementMessage($form);
      $numberElement = NULL;
    }

    $form['number_element_settings']['number_element_key'] = [
      '#type' => 'select',
      '#title' => $this->t('Number element'),
      '#options' => $this->getNumberElements(),
      '#default_value' => $numberElement,
      '#required' => TRUE,
      '#empty_option' => (empty($numberElement)) ? $this->t('- Select -') : NULL,
    ];

    $form = parent::buildConfigurationForm($form, $form_state);
    $form['element_settings']['#weight'] = -2;

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function alterForm(array &$form, FormStateInterface $form_state, WebformSubmissionInterface $webform_submission) {
    parent::alterForm($form, $form_state, $webform_submission);

    if (!isset($form['elements'][$this->configuration['element_key']])) {
      return;
    }
    $optionsElement = &$form['elements'][$this->configuration['element_key']];

    if (!isset($form['elements'][$this->configuration['number_element_key']])) {
      return;
    }
    $numberElement = &$form['elements'][$this->configuration['number_element_key']];

    $chosenOptions = [];

    if (!empty($form_state->getUserInput()[$this->configuration['element_key']])) {
      $chosenOptions = $form_state->getUserInput()[$this->configuration['element_key']];
    }
    elseif (!empty($form_state->getValue($this->configuration['element_key']))) {
      $chosenOptions = $form_state->getValue($this->configuration['element_key']);
    }

    if (!empty($chosenOptions) && empty($optionsElement['#webform_multiple'])) {
      $chosenOptions = [$chosenOptions];
    }

    if ($chosenOptions !== []) {
      $limits = $this->getOptionsLimits();
      $remaining = array_reduce(
        $chosenOptions,
        function (int $remaining, $option) use ($limits) {
          return min($remaining, $limits[$option]['remaining']);
        },
        $numberElement['#max'] ?? PHP_INT_MAX
      );

      if ($remaining < PHP_INT_MAX) {
        // Set #max on the number element to make sure this submission won't
        // surpass the limit.
        $numberElement['#max'] = $remaining;

        // Update the number element value if it's greater than the limit.
        $value = $this->formState->getUserInput()[$this->configuration['number_element_key']]
          ?? $this->formState->getValue($this->configuration['number_element_key']);
        if ($value > $remaining) {
          $numberElement['#value'] = $remaining;
        }
      }
    }

    // Add ajax to update the number element.
    $id = 'webform-options-number-limit';
    static::buildAjaxElementWrapper($id, $form['elements']);
    static::buildAjaxElementTrigger($id, $optionsElement);
    static::buildAjaxElementTrigger($id, $numberElement);
    static::buildAjaxElementUpdate($id, $form['elements']);
  }

  /**
   * {@inheritdoc}
   */
  protected function getLimitInformation($label, $limit, $total) {
    $total = $total ?: 0;
    $numberValue = 0;

    if (!empty($this->formState->getUserInput()[$this->configuration['number_element_key']])) {
      $numberValue = $this->formState->getUserInput()[$this->configuration['number_element_key']];
    }
    elseif (!empty($this->formState->getValue($this->configuration['number_element_key']))) {
      $numberValue = $this->formState->getValue($this->configuration['number_element_key']);
    }
    $actualRemaining = ($limit) ? $limit - $total - ($numberValue - 1) : 0;
    $displayedRemaining = ($limit) ? $limit - $total : 0;

    if (empty($limit) && $limit !== 0) {
      $status = WebformOptionsLimitHandlerInterface::LIMIT_STATUS_UNLIMITED;
    }
    elseif ($actualRemaining <= 0) {
      $status = WebformOptionsLimitHandlerInterface::LIMIT_STATUS_NONE;
    }
    elseif ($actualRemaining === 1) {
      $status = WebformOptionsLimitHandlerInterface::LIMIT_STATUS_SINGLE;
    }
    else {
      $status = WebformOptionsLimitHandlerInterface::LIMIT_STATUS_MULTIPLE;
    }

    return [
      'label' => $label,
      'limit' => $limit,
      'total' => $total,
      'remaining' => $displayedRemaining,
      'status' => $status,
    ];
  }

  /**
   * {@inheritdoc}
   */
  protected function getOptionsTotals(array $values = []) {
    $totals = [];
    $options = $this->getElementOptions();

    foreach ($options as $key => $option) {
      $numberTotal = $this->getNumberElementTotal($this->configuration['element_key'], $key);
      $totals[$key] = $numberTotal;
    }

    return $totals;
  }

  /**
   * Get the total number of submissions for the number element.
   */
  protected function getNumberElementTotal(string $elementName, string $elementOption): int {
    $webform = $this->getWebform();

    /** @var \Drupal\Core\Database\StatementInterface $result */
    $query = $this->database->select('webform_submission', 's');
    $query->join('webform_submission_data', 'sd', 's.sid = sd.sid');
    $query->join('webform_submission_data', 'sda', 'sd.sid = sda.sid AND sda.name = :element_name AND sda.value = :element_option', [
      ':element_name' => $elementName,
      ':element_option' => $elementOption,
    ]);
    $query->addExpression('SUM(sd.value)', 'sum');
    $query->condition('sd.name', $this->configuration['number_element_key']);
    $query->condition('sd.webform_id', $webform->id());

    return (int) $query->execute()->fetchField();
  }

}
